//
//  MiscMergeTemplate.m
//            Written by Don Yacktman and Carl Lindberg
//        Copyright 1998 by Don Yacktman and Carl Lindberg.
//                     All rights reserved.
//      This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import "MiscMergeTemplate.h"
#import <Foundation/NSCharacterSet.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSUtilities.h>
#import <Foundation/NSAutoreleasePool.h>
#import <Foundation/NSObjCRuntime.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "NSScanner+MiscMerge.h"
#import "MiscMergeCommand.h"
#import "MiscMergeCommandBlock.h"
#import "_MiscMergeFieldCommand.h"
#import "_MiscMergeCopyCommand.h"
#import "_MiscMergeDelayedParseCommand.h"

@implementation MiscMergeTemplate
/*"
 * This class contains the template that is used by a merge engine. It
 * performs two functions:  (1) parse a string or text file into the
 * commands required by a merge ending and (2) act as a container for the
 * commands once they have been parsed, providing them to a merge engine as
 * needed.
 * 
 * Typically, MiscMergeTemplate objects are used in a very simple way: they
 * are instantiated, given the ASCII text or string to parse, and then
 * passed to MiscMergeEngine instances as needed.  That's it!
 * 
 * It should be noted that template text which is simply copied from the
 * template into the merged output (i.e. any text outside of a merge
 * command) is actually turned into a special "copy" command by the parsing
 * algorithm. This allows the merge engine to deal exclusively with
 * MiscMergeCommand subclasses to perform a merge.  This implementation
 * detail should not affect anything that would normally be done with this
 * object, but it is important to understand this fact if attempting to
 * understand the data structure created by the parsing routines.
 * 
 * If a command string contains merge commands inside itself, then a
 * special "delayed command" class will used.  That class will, during a
 * merge, create an engine, perform a merge on its text, and then parse
 * itself into the correct type of command.  This allows merges to contain
 * commands that change depending upon the data records.
 * 
 * Commands created while parsing are always added to the "current" command
 * block.  By using -#pushCommandBlock: and -#popCommandBlock,
 * MiscMergeCommand subclasses can temporarily substitute their own command
 * block to be the "current" block during the parsing process.  This way,
 * container-type commands such as if clauses and loops can know which
 * commands they hold.
"*/

/*"
 * Returns the default string used to start a merge command, "".  A
 * subclass of MiscMergeTemplate could override this method.
"*/
+ (NSString *)defaultStartDelimiter
{
//	return @"(";
//	return @"";
    /* This works better for whatever reason. Due to some unknown pecularities,
       a constant NSString doesn't work under Windows with Apple's
       implementation. */
	return [NSString stringWithCString:""];
}

/*"
 * Returns the default string used to end a merge command, "".  A
 * subclass of MiscMergeTemplate could override this method.
"*/
+ (NSString *)defaultEndDelimiter
{
//	return @")";
//	return @"";
	/* This works better than a constant NSString for whatever reason.  See above. */
	return [NSString stringWithCString:""];
}

/*" Creates a new, autoreleased MiscMergeTemplate. "*/
+ template
{
	return [[[self alloc] init] autorelease];
}

/*"
 * Creates a new, autoreleased MiscMergeTemplate, and parses aString.
"*/
+ templateWithString:(NSString *)aString
{
	return [[[self alloc] initWithString:aString] autorelease];
}

/*"
 * Initializes the MiscMergeTemplate instance and returns self.  This is
 * the designated initializer.  The start and end delimiters are set to the
 * values returned by +#defaultStartDelimiter and +#defaultEndDelimiter.
"*/
- init
{
	[super init];
	topLevelCommands = [[MiscMergeCommandBlock alloc] init];
	commandStack = [[NSMutableArray arrayWithObject:topLevelCommands] retain];
	startDelimiter = [[[self class] defaultStartDelimiter] retain];
	endDelimiter = [[[self class] defaultEndDelimiter] retain];
	return self;
}

/*"
 * Initializes the MiscMergeTemplate, then parses string.
"*/
- initWithString:(NSString *)string
{
	[self init];
	[self parseString:string];
	return self;
}

/*"
 * Loads the contents of filename, then calls -#initWithString:.
"*/
- initWithContentsOfFile:(NSString *)filename
{
	NSString *fileString = [[[NSString alloc] initWithContentsOfFile:filename] autorelease];
	if (fileString == nil) NSLog(@"%@: Could not read template file %@", [self class], filename);
	return [self initWithString:fileString];
}

- (void)dealloc
{
	[commandStack release];
	[topLevelCommands release];
    [startDelimiter release];
    [endDelimiter release];
	[_parseStopChars release];
    [_filename release];
	[super dealloc];
}

/*" Returns the string used to start a merge command. "*/
- (NSString *)startDelimiter
{
	return startDelimiter;
}

/*" Returns the string used to end a merge command. "*/
- (NSString *)endDelimiter
{
	return endDelimiter;
}


/*"
 * Sets the strings used to start and end a merge command. %{startDelim}
 * and %{endDelim} can be the same strings, but if they are the same (or
 * one is a prefix of the other), then there will be no ability to have
 * nested commands, so the "delayed" feature will not be available.
"*/
- (void)setStartDelimiter:(NSString *)startDelim endDelimiter:(NSString *)endDelim
{
	[startDelimiter autorelease];
    [endDelimiter   autorelease];
    startDelimiter = [startDelim retain];
    endDelimiter   = [endDelim retain];
	[_parseStopChars release];
	_parseStopChars = nil;
}

/*"
 * Returns the filename that was used to create the template, or nil if not
 * parsed from a file.  Used for reporting errors.
"*/
- (NSString *)filename
{
    return _filename;
}

/*"
 * Sets the filename the template was created from, which is used in
 * reporting errors.  This is normally set by -#parseContentsOfFile:, but
 * this method can be used if -#parseString: needs to be called directly
 * but the original source did come from a file.
"*/
- (void)setFilename:(NSString *)filename
{
    [_filename autorelease];
    _filename = [filename retain];
}

/*"
 * Pushes aBlock on the command stack.  aBlock becomes the current command
 * block until popped off (or another block is placed on top of it).
"*/
- (void)pushCommandBlock:(MiscMergeCommandBlock *)aBlock
{
	[commandStack addObject:aBlock];
}

/*"
 * Pops the command block aBlock off of the command stack, so the previous
 * command block will again be the "current" command block.  If aBlock is
 * not at the top of the command stack, logs an error and does nothing.
 * Basically the same as -#popCommandBlock except it does the extra sanity
 * check.
"*/
- (void)popCommandBlock:(MiscMergeCommandBlock *)aBlock
{
	if (aBlock && [commandStack lastObject] != aBlock)
	{
		[self reportParseError:@"Error, command stack mismatch"];
		return;
	}
	
	[self popCommandBlock];
}

/*"
 * Pops the top command block off the command stack, so the previous
 * command block will again be the "current" command block.
"*/
- (void)popCommandBlock
{
	if ([commandStack count] <= 1)
	{
        [self reportParseError:@"Error, cannot pop last command block"];
		return;
	}

	[commandStack removeLastObject];
}

/*"
 * Returns the "current" command block, i.e. the command block at the top
 * of the command stack.  As they are parsed, commands are always added to
 * the command block returned by this method.
"*/
- (MiscMergeCommandBlock *)currentCommandBlock
{
	return [commandStack lastObject];
}

/*"
 * Returns the "top level" command block of the MiscMergeTemplate, which is
 * basically the series of commands to be executed to generate the merge
 * file.  The top level block is always at the bottom of the command stack.
"*/
- (MiscMergeCommandBlock *)topLevelCommandBlock
{
	return topLevelCommands;
}

/*"
 * Given the command string %{aCommand}, this method determines which
 * MiscMergeCommand subclass implements the merge command.  It returns the
 * class object needed to create instances of the MiscMergeCommand
 * subclass.
 * 
 * This method works by asking the runtime if it can find Objective-C
 * classes with specific names.  The name that is looked up is build from
 * the first word found in %{aCommand}.  The first word is turned to all
 * lower case, with the first letter upper case, and then sandwiched
 * between "Merge" and "Command".  For example, the merge command "if xxx
 * = y" has the word "if" as the first word.  Thus, the class
 * "MergeIfCommand" will be searched for. If the desired class cannot be
 * found, then it is assumed that the merge command is giving the name of a
 * field which should be inserted into the output document.
 * 
 * To avoid name space conflicts, all internal merge commands actually use
 * a slightly different name.  Thus, there really is no "MergeIfCommand" to
 * be found.  This method, when it doesn't find the "MergeIfCommand" class,
 * will search for another class, with a private name.  That class will be
 * found. (If it wasn't found, then the default "field" command class would
 * be returned.)  This allows a programmer to override any built in
 * command. To override the "if" command, simply create a "MergeIfCommand"
 * class and it will be found before the built in class.  If a programmer
 * wishes to make a particular command, such as "omit", inoperative, this
 * technique may be used to override with a MiscMergeCommand subclass that
 * does nothing.
"*/
- (Class)classForCommand:(NSString *)aCommand
{
	NSString *realCommand = [[aCommand firstWord] capitalizedString];
	NSString *className;
	Class theClass;
	
	className = [NSString stringWithFormat:@"MiscMerge%@Command", realCommand];
	theClass = NSClassFromString(className);
	
	if (theClass == Nil)
	{
		className = [NSString stringWithFormat:@"_MiscMerge%@Command", realCommand];
        theClass = NSClassFromString(className);
	}
	
	if (theClass == Nil)
	{
		theClass = [_MiscMergeFieldCommand class];
	}
	
	return theClass;
}

/*"
 * Reports the given error message, prefixed with filename and current line
 * number.
"*/
- (void)reportParseError:(NSString *)format, ...
{
    NSString *errorMessage;
    va_list args;

    va_start(args, format);
    errorMessage = [[NSString alloc] initWithFormat:format arguments:args];
    va_end(args);

    if (_filename != nil)
        NSLog(@"%@: %d: %@", _filename, _lineNumber, errorMessage);
    else
        NSLog(@"Line %d: %@", _lineNumber, errorMessage);

    [errorMessage release];
}

- (void)_addCommand:(MiscMergeCommand *)command
{
	[[self currentCommandBlock] addCommand:command];
}

- (void)_addBetweenString:(NSString *)betweenString
{
	id command = [[_MiscMergeCopyCommand alloc] init];
	[command parseFromRawString:betweenString];
	[self _addCommand:command];
	[command release];
    _lineNumber += [betweenString numOfString:@"\n"];
}

- (void)_addCommandString:(NSString *)commandString
{
	Class commandClass = [self classForCommand:commandString];
	id command = [[commandClass alloc] init];
	[self _addCommand:command];
	[command parseFromString:commandString template:self];
	[command release];
    _lineNumber += [commandString numOfString:@"\n"];
}

/*"
 * Loads the contents of filename and calls -#parseString:.
"*/
- (void)parseContentsOfFile:(NSString *)filename
{
        NSString *string = [[NSString alloc] initWithContentsOfFile:filename];
	if (string == nil) NSLog(@"%@: Could not read template file %@", [self class], filename);
	[self setFilename:filename];
	[self parseString:string];
	[string release];
}

/*"
 * Parses the template in %{string}.
"*/
- (void)parseString:(NSString *)string
{
	NSAutoreleasePool *localPool = [[NSAutoreleasePool alloc] init];
	NSScanner *scanner = [NSScanner scannerWithString:string];
	NSMutableString *accumString = [[NSMutableString alloc] init];
	NSString *startDelimiterString = [self startDelimiter];
	NSString *endDelimiterString = [self endDelimiter];
	NSString *currString;
	int nestingLevel = 0;
	int maxNestingLevel = 0;
	BOOL inQuotes = NO;


	if (_parseStopChars == nil)
	{
		NSString *charString = [NSString stringWithFormat:@"\"\\%@%@",
            [startDelimiterString substringToIndex:1],
            [endDelimiterString substringToIndex:1]];
		_parseStopChars = [NSCharacterSet characterSetWithCharactersInString:charString];
		[_parseStopChars retain];
	}
	
	[scanner setCharactersToBeSkipped:nil];
	_lineNumber = 1;

	// may want to flush localPool every 50 loops or so...
	while (![scanner isAtEnd])
	{
	  if ([scanner scanUpToCharactersFromSet:_parseStopChars intoString:&currString])
	    [accumString appendString:currString];

        if ([scanner scanString:@"\\"])
        {
            BOOL     foundDelimiter;
            NSString *delimiter = nil;

            /* Look for the delimiters.  As a side effect, move scanner along */
            foundDelimiter = [scanner scanString:startDelimiterString intoString:&delimiter] ||
                             [scanner scanString:endDelimiterString intoString:&delimiter];
 
           /*
            * Leave the backslash if we are not quoting a delimiter string
            * -- we would wreak havoc on RTF template files otherwise.
            * Also leave the backslash if we are deeply nesting or inside
            * quotes, as those sections will get reduced later.
            */
            if (nestingLevel > 1 || (nestingLevel == 1 && inQuotes) || !foundDelimiter)
            {
                [accumString appendString:@"\\"];
            }

            if (foundDelimiter)
                [accumString appendString:delimiter];
        }
        else if ([scanner scanString:@"\""])
        {
            [accumString appendString:@"\""];
            if (nestingLevel == 1) inQuotes = !inQuotes;
        }
        else if (nestingLevel > 0 && [scanner scanString:endDelimiterString])
        {
            if (nestingLevel > 1)
            {
                [accumString appendString:endDelimiterString];
            }
            else
            {
                /* Special hack for the delayed parsing stuff. Hm. */
                if (maxNestingLevel > 1)
                {
                    id command = [[_MiscMergeDelayedParseCommand alloc] init];
                    _lineNumber += [accumString numOfString:@"\n"];
                    [command parseFromString:accumString template:self];
                    [self _addCommand:command];
                    [command release];
                }
                else
                {
                    if ([accumString length] > 0)
                        [self _addCommandString:[[accumString copy] autorelease]];
                }
                [accumString setString:@""];
                inQuotes = NO;
                maxNestingLevel = 0;
            }

            nestingLevel--;
        }
        else if ([scanner scanString:startDelimiterString])
        {
            if (nestingLevel > 0)
            {
                [accumString appendString:startDelimiterString];
            }
            else
            {
                if ([accumString length] > 0)
                    [self _addBetweenString:[[accumString copy] autorelease]];
                [accumString setString:@""];
            }

            nestingLevel++;
            if (nestingLevel > maxNestingLevel) maxNestingLevel = nestingLevel;
        }
        else if ([scanner scanLetterIntoString:&currString])
        {
            // If we can scan the end delimiter, it's an error, otherwise
            // it's just a string that starts with the same char as a
            // delimiter.
            [accumString appendString:currString];
        }
    }
	
	if ([accumString length] > 0)
	{
		[self _addBetweenString:[[accumString copy] autorelease]];
	}
	
	[localPool release];
}

@end
